"""@package URLs

    A Python3 class that allows access to all of the functionality in the
    Ally/TradeKing API.

    This package attempts to stay on top of changes to the API and allow an
    easy to user iterface with the Ally Invest API. The API does no formatting
    for the user. A response format of 'xml' or 'json' can be specified and
    the API responses will be returned as the raw XML or JSON, respectively.

    This API was built with the developer in mind and should allow a developer
    to build applications around the Ally Invest API without having to deal with
    accessing and managing the requests and responses.

    This project was inspired my PyAlly (https://github.com/alienbrett/PyAlly).
"""
class URLs:
    """ The URLs class will handle all of the URLs. The purpose of this class is
        to essentially store and serve all of the URL strings useful to the
        Ally Invest API.

        There is no processing of the URLs or the URL parameters done in this class
        all of that logic is handled in the AllyAPI class.
    """
    def __init__(self, response_format="json"):
        """The URLs class constructor which defines all of the URLs used by the API.

            When adding new API functionality the URL needs to be added here.
            Examples abound of the format used by this implementation of the API.

            @param self - the object pointer
            @param response_format - format of the response. Valid values are 'xml' and 'json'.
                Specifying 'xml' will return an ElementTree containing the response XML while
                'json' will return the response in the JSON format.
        """
        self.format = response_format

        self.base_url = "https://api.tradeking.com/v1/"
        # self.request_token = "https://developers.tradeking.com/oauth/request_token"
        # self.user_auth = "https://developers.tradeking.com/oauth/authorize"
        # self.resource_owner_key = "https://developers.tradeking.com/oauth/resource_owner_key"

        # account
        self.accounts = "accounts.{format}".format(format=self.format)
        self.accounts_balances = "accounts/balances.{format}".format(format=self.format)
        self.account = "accounts/{id}.{format}".format(format=self.format, id="{id}")
        self.account_balances = "accounts/{id}/balances.{format}".format(format=self.format, id="{id}")
        self.account_history = "accounts/{id}/history.{format}".format(format=self.format, id="{id}")
        self.account_holdings = "accounts/{id}/holdings.{format}".format(format=self.format, id="{id}")

        # market
        self.clock = "market/clock.{format}".format(format=self.format)
        self.quote = "market/ext/quotes.{format}".format(format=self.format)
        self.news_search = "/market/news/search.{format}".format(format=self.format)
        self.news_article = "market/news/{article_id}.{format}".format(format=self.format, article_id="{article_id}")
        self.toplists = "market/toplists/{listtype}.{format}".format(format=self.format, listtype="{listtype}")

        # member
        self.member_profile = "member/profile.{format}".format(format=self.format)

        # Utilities
        self.status = "utility/status.{format}".format(format=self.format)
        self.version = "utility/version.{format}".format(format=self.format)

        # watchlists
        self.watchlists = "watchlists.{format}".format(format=self.format)

    def base_url(self):
        """Returns the API request endpoint.
            @param self - the object pointer
        """
        return self.base_url

    # def request_token(self):
    #     """
    #         @param self - the object pointer
    #     """
    #     return self.request_token
    #
    # def user_auth(self):
    #     """
    #         @param self - the object pointer
    #     """
    #     return self.user_auth
    #
    # def resource_owner_key(self):
    #     """
    #         @param self - the object pointer
    #     """
    #     return self.resource_owner_key

    """
        Accounts
    """
    def accounts_url(self):
        """Combines the request endpoint and accounts API URLs
            @param self - the object pointer
        """
        return self.base_url + self.accounts

    def accounts_balances_url(self):
        """Combines the request endpoint and accounts balances API URLs
            @param self - the object pointer
        """
        return self.base_url + self.accounts_balances

    def account_url(self):
        """Combines the request endpoint and account API URLs
            @param self - the object pointer
        """
        return self.base_url + self.account

    def account_balances_url(self):
        """Combines the request endpoint and account balances API URLs
            @param self - the object pointer
        """
        return self.base_url + self.account_balances

    def account_history_url(self):
        """Combines the request endpoint and account history API URLs
            @param self - the object pointer
        """
        return self.base_url + self.account_history

    def account_holdings_url(self):
        """Combines the request endpoint and account holding API URLs
            @param self - the object pointer
        """
        return self.base_url + self.account_holdings

    """
        ORDER/TRADE
        TODO:
            GET accounts/:id/orders
            POST accounts/:id/orders
            POST accounts/:id/orders/preview
    """

    """
        Market
        TODO:
            GET market/options/search
            GET market/options/strikes
            GET market/options/expirations
            GET market/timesales
    """
    def clock_url(self):
        """Combines the request endpoint and market clock API URLs
            @param self - the object pointer
        """
        return self.base_url + self.clock

    def quote_url(self):
        """Combines the request endpoint and quote API URLs
            @param self - the object pointer
        """
        return self.base_url + self.quote

    def news_search_url(self):
        """Combines the request endpoint and news search API URLs
            @param self - the object pointer
        """
        return self.base_url + self.news_search

    def news_article_url(self):
        """Combines the request endpoint and news article API URLs
            @param self - the object pointer
        """
        return self.base_url + self.news_article

    def toplists_url(self):
        """Combines the request endpoint and toplists API URLs
            @param self - the object pointer
        """
        return self.base_url + self.toplists

    """
        Member
    """
    def member_profile_url(self):
        """Combines the request endpoint and member profile API URLs
            @param self - the object pointer
        """
        return self.base_url + self.member_profile

    """
        Utilities
    """
    def status_url(self):
        """Combines the request endpoint and server status API URLs
            @param self - the object pointer
        """
        return self.base_url + self.status

    def version_url(self):
        """Combines the request endpoint and API version API URLs
            @param self - the object pointer
        """
        return self.base_url + self.version

    """
        WATCHLIST
        TODO:
            GET watchlists/:id
            DELETE watchlists/:id
            POST watchlists/:id/symbols
            DELETE watchlists/:id/symbols
    """
    def get_watchlists_url(self):
        """Combines the request endpoint and watchlist get URLs. Note this is the
        same URL as the POST call (URLs.post_watchlist_url()).
        """
        return self.base_url + self.watchlists

    def post_watchlist_url(self):
        """Combines the request endpoint and watchlist post URLs"""
        return self.base_url + self.watchlists

    """
        STREAMING OPERATIONS
            MARKET
            TODO:
                GET market/quotes
    """



from .response import *

class AccountBalancesResponse(Response):
    def __init__(self, account_id, response_format, data):
        super().__init__(response_format, data)
        self.account_id = account_id

    def __parse_xml(self, data):
        pass

    def __parse_json(self, data):
        pass



from .response import *
from .quote import *

class QuotesResponse(Response):
    def __init__(self, response_format, data):
        super().__init__(response_format, data)
        self.quotes = []
        if response_format.lower() == 'xml':
            self.__parse_xml(data)
        elif response_format.lower() == 'json':
            self.__parse_json(data)

    def __parse_xml(self, data):
        pass

    def __parse_json(self, data):
        super().parse_json(data)
        if isinstance(self.json['quotes']['quote'], list):
            for quote_json in self.json['quotes']['quote']:
                quote = Quote()
                quote.from_json(quote_json)
                self.quotes.append(quote)
        else:
            quote = Quote()
            quote.from_json(self.json['quotes']['quote'])
            self.quotes.append(quote)

    def get_quotes(self):
        return self.quotes



"""@package AllyAPI
    A Python3 class that allows access to all of the functionality in the
    Ally/TradeKing API.

    This package attempts to stay on top of changes to the API and allow an
    easy to user iterface with the Ally Invest API. The API does no formatting
    for the user. A response format of 'xml' or 'json' can be specified and
    the API responses will be returned as the raw XML or JSON, respectively.

    This API was built with the developer in mind and should allow a developer
    to build applications around the Ally Invest API without having to deal with
    accessing and managing the requests and responses.

    This project was inspired my PyAlly (https://github.com/alienbrett/PyAlly).
"""

from requests_oauthlib import OAuth1
from xml.etree import ElementTree
import datetime
import requests
import json

from .URLs import URLs

class AllyAPI:
    """The AllyAPI class providing blackbox use of the Ally Invest API.

    This is the main class of the API module. This should be the only class used in
    applictions built around the API. The AllyAPI class allows access to the GET and
    POST requests supported by Ally Invest.

    Missing Functionality:
        ORDER/TRADE
            GET accounts/:id/orders
            POST accounts/:id/orders
            POST accounts/:id/orders/preview
        MARKET
            GET market/options/search
            GET market/options/strikes
            GET market/options/expirations
            GET market/timesales
        WATCHLIST
            GET watchlists/:id
            DELETE watchlists/:id
            POST watchlists/:id/symbols
            DELETE watchlists/:id/symbols
        STREAMING OPERATIONS
            MARKET
                GET market/quotes
    """
    def __init__(self, oauth_secret, oauth_token, client_key,
                response_format="json"):
        """AllyAPI constructor. Sets the response format on all of the URLs and
            the oauth/client keys required to access the API.

            Parameters
                @param self - the object pointer
                @param oauth_secret - secret oauth key from Ally
                @param oauth_token - oauth token from Ally
                @param client_key - client key from Ally
                @param response_format - format of the response. Valid values are 'xml' and 'json'.
                    Specifying 'xml' will return an ElementTree containing the response XML while
                    'json' will return the response in the JSON format.
        """
        self.format = response_format
        self.url = URLs(response_format=response_format)

        self.oauth_secret = oauth_secret
        self.oauth_token = oauth_token
        self.client_key = client_key
        self.client_secret = client_key

        self.auth_time = None
        self.auth = None
        self.valid_auth_dt = datetime.timedelta(seconds=10)

    def __create_auth(self):
        """A private method to create the OAuth1 object, if necessary."""
        now = datetime.datetime.now()

        if self.auth == None or self.auth_time + self.valid_auth_dt < now:
            self.auth_time = now
            self.auth = OAuth1(self.client_key, self.client_secret, self.oauth_token,
                          self.oauth_secret, signature_type='auth_header')

    def __get_symbol_string(self, symbols):
        """Returns a string that is either a single quote or a comma-separated
            list of quotes depending on the type of quotes.
            @param self - the object pointer
            @param symbols - single ticker or list of ticker to get quotes for
        """
        if not isinstance(symbols, str): # list
            symbols = ",".join(symbols)
        return symbols

    def __to_format(self, response):
        """A private method to return the API response in the desired format
            @param self - the object pointer
            @param response - response from the Ally Invest API
        """
        if self.format == "json":
            return response.json()
        else:
            return ElementTree.fromstring(response.content)

    def __get_data(self, url):
        """A private method to return the requested data in the requested format
            for a given URL.
            @param self - the object pointer
            @param url - API URL to access
        """
        self.__create_auth()
        return self.__to_format(requests.get(url, auth=self.auth))

    def __submit_post(self, url, data):
        """A private method to submit a post request to the Ally Invest server
            @param self - the object pointer
            @param url - API URL to access
            @param data - payload for the HTTP request
        """
        self.__create_auth()
        return self.__to_format(requests.post(url, data=data, auth=self.auth))

    def get_accounts(self):
        """Returns all of the user's accounts."""
        return self.__get_data(self.url.accounts_url())

    def get_accounts_balances(self):
        """Returns the balances of all of the user's accounts."""
        return self.__get_data(self.url.accounts_balances_url())

    def get_account(self, id):
        """Returns a specific account provided the account ID (account number)
            @param self - the object pointer
            @param id - account number
        """
        return self.__get_data(self.url.account_url().format(id=str(id)))

    def get_account_balances(self, id):
        """Returns the balances of a specific account (ID = account number)
            @param self - the object pointer
            @param id - account number
        """
        return self.__get_data(self.url.account_balances_url().format(id=str(id)))

    def get_account_history(self, id):
        """Returns the history of a specific account (ID = account number)
            @param self - the object pointer
            @param id - account number
        """
        return self.__get_data(self.url.account_history_url().format(id=str(id)))

    def get_account_holdings(self, id):
        """Returns the holdings of a specific account (ID = account number)
            @param self - the object pointer
            @param id - account number
        """
        return self.__get_data(self.url.account_holdings_url().format(id=str(id)))

    def get_market_clock(self):
        """Returns the state of the market, the time until next state change,
            and current server timestamp.
            @param self - the object pointer
        """
        return self.__get_data(self.url.clock_url())

    def get_quote(self, symbols):
        """Returns quote information for a single ticker or list of tickers.
            Note: this function does not implement selecting customer FIDs as
            described in the API documentation. These can be filtered from the return
            if need be.
            @param self - the object pointer
            @param symbols - single ticker or list of ticker to get quotes for
        """
        url = self.url.quote_url()+"?symbols={symbols}"
        symbols = self.__get_symbol_string(symbols)
        return self.__get_data(url.format(symbols=symbols))

    def news_search(self, symbols, startdate=None, enddate=None, maxhits=10):
        """Retrieves a listing of news headlines based on symbols.
            @param self - the object pointer
            @param symbols - single ticker or list of ticker to get quotes for
            @param startdate - search for articles between this date and enddate
            @param enddate - search for articles between this date and startdate
            @param maxhits - number of articles to return
        """
        if startdate is None or enddate is None:
            print("news_search: either enddate or startdate is not specified, ignoring both.")

        if (startdate is not None and enddate is not None) and (enddate < startdate):
            print("news_search: start date is after end date.")
            raise Exception("Start date is after end date in news search.")

        url = self.url.news_search_url() + "?symbols={syms}&maxhits={mxhits}".format(mxhits=maxhits, syms="{syms}")
        if startdate is not None and enddate is not None:
            url += "&startdate={sdate}&enddate={edate}" \
                .format(sdate=startdate.strftime("%m/%d/%Y"), edate=enddate.strftime("%m/%d/%Y"))

        symbols = self.__get_symbol_string(symbols)
        return self.__get_data(url.format(syms=symbols))

    def get_news_article(self, article_id):
        """Gets a single news article based on the article ID. This ID can be retrieved
            from the news_search()  function.
            @param self - the object pointer
            @param article_id - ID of the article to retrieve
        """
        return self.__get_data(self.url.news_article_url().format(article_id=article_id))

    def get_toplists(self, listtype="topgainers", exchange="N"):
        """Returns a ranked list depending on listtype and exchange.
            @param listtype - type of list to be queried, accepted values are:
                'toplosers': top losers by dollar amount
                'toppctlosers': top percentage losers
                'topvolume': top volume
                'topactive': top active
                'topgainers': top gainers by dollar amount (default)
                'toppctgainers': top percentage gainers
            @param exchange - exchange to be queried, accepted values are:
                'A': American Stock Exchange
                'N': New York Stock Exchange (default)
                'Q': NASDAQ
                'U': NASDAQ Bulletin Board
                'V': NASDAQ OTC Other
        """
        url = self.url.toplists_url().format(listtype=listtype)
        url += "?exchange={ex}".format(ex=exchange)
        return self.__get_data(url)

    def get_member_profile(self):
        """Returns general information associated with the user including account
            numbers and account information.
            @param self - the object pointer
        """
        return self.__get_data(self.url.member_profile_url())

    def get_status(self):
        """Returns an error if the API endpoint/server is unavailable. Otherwise
            returns the current server timestamp.
            @param self - the object pointer
        """
        return self.__get_data(self.url.status_url())

    def get_version(self):
        """Gets the current version of the API of the endpoint called.
            @param self - the object pointer
        """
        return self.__get_data(self.url.version_url())

    def get_watchlists(self):
        """Retrieves all watchlists belonging to the member.
            @param self - the object pointer
        """
        return self.__get_data(self.url.get_watchlists_url())


    def create_watchlist(self, watchlist_name, symbols=""):
        """Creates a watchlist and adds a symbol or list of symbols to a watchlist.
            WARNING: There appears to be an issue when adding a list of symbols.
                It is recommended that one ticker symbol is added at a time.
            @param self - the object pointer
            @param watchist_id - name of the watchlist
            @param symbols - single ticker or list of tickers to add to the watchlist
        """
        print("WARNING create_watchlist(): There appears to be an issue when adding a list of symbols. It is recommended that one ticker symbol is added at a time.")
        payload = {"id": watchlist_name}
        if not symbols == "":
            payload["symbols"] = symbols
        return self.__submit_post(self.url.post_watchlist_url(), payload)



import json

class Response():
    def __init__(self, response_format, data):
        self.response_format = response_format
        self.raw_data = data

    def parse_xml(self, data):
        pass

    def parse_json(self, data):
        self.json = json.loads(json.dumps(data["response"]))

    def get_raw_data(self):
        return self.raw_data



from .account_balances import *
from .accounts_balances import *
from .account import *          # accounts/:id
from .accounts import *         # accounts/
from .quotes import *           # reponse from executing QuotesRequest
from .account_holdings import *



from .request import *
from ..responses.accounts_balances import *

class AccountsBalancesRequest(Request):
    def __init__(self, response_format='json'):
        super().__init__(response_format)

    def execute(self, ally_api):
        return AccountsBalancesResponse(self.response_format, ally_api.get_accounts_balances())



from .response import *
from .holding import *

class AccountHoldingsResponse(Response):
    def __init__(self, account_id, response_format, data):
        super().__init__(response_format, data)
        self.account_id = account_id
        self.holdings = []
        if response_format.lower() == 'xml':
            self.__parse_xml(data)
        elif response_format.lower() == 'json':
            self.__parse_json(data)

    def __parse_xml(self, data):
        super().parse_json(data)
        for holding_json in data['accountholdings']:
            holding = Holding()
            self.holdings.append(holding.from_xml(holding_json))

    def __parse_json(self, data):
        super().parse_json(data)
        for holding_json in self.json['accountholdings']['holding']:
            holding = Holding()
            holding.from_json(holding_json)
            self.holdings.append(holding)

    def get_holdings(self):
        return self.holdings



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='AllyInvestPy',
    version='1.0.2',
    description='A blackbox Ally Invest/TradeKing API interface for application developers.',
    url='https://github.com/anthonymorast/AllyInvest.py',
    author='Anthony Morast',
    author_email='anthony.a.morast@gmail.com',
    license='MIT',
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=[
        'requests',
        'requests_oauthlib',
    ]
)



from .request import *
from ..responses.account_balances import *

class AccountBalancesRequest(Request):
    def __init__(self, account_id, response_format='json'):
        super().__init__(response_format)
        self.account_id = account_id

    def execute(self, ally_api):
        return AccountBalancesResponse(self.account_id, self.response_format, ally_api.get_account_balances(self.account_id))
